"""
Install dependencies and copies includes into a zipped file ready for deployment.

The resulting zip file is structured correctly to be deployed to the lambda
function/layer target via an S3 upload and subsequent publish command.
"""
import argparse
import pathlib
import shutil
import typing

from reviser import bundling
from reviser import definitions
from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return ["--reinstall"]


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    parser.add_argument(
        "--reinstall",
        action="store_true",
        help="""
            Add this flag to reinstall dependencies on a repeated
            bundle operation. By default, dependencies will remain
            cached for the lifetime of the shell to speed up the
            bundling process. This will force dependencies to be
            installed even if they had been installed previously.
            """,
    )
    parser.add_argument(
        "--output",
        "-o",
        help="Output the bundled artifacts into the specified output path.",
    )


def _copy_to_output(
    targets: typing.List["definitions.Target"], output_dir: typing.Optional[str]
):
    """Copy the outputs of the targets into the specified output directory."""
    if not output_dir:
        return
    output_path = pathlib.Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)
    for target in targets:
        for name in target.names:
            shutil.copy(
                target.bundle_zip_path,
                output_path.joinpath(f"{name}-{target.kind.value}.zip"),
            )


def run(ex: "interactivity.Execution"):
    """Execute a bundle operation on the selected function/layer targets."""
    bundled_targets = bundling.create(
        context=ex.shell.context,
        selection=ex.shell.selection,
        reinstall=ex.args.get("reinstall", False),
    )
    _copy_to_output(bundled_targets.targets, ex.args.get("output"))
    print("\n\n")
    return ex.finalize(
        status="BUNDLED",
        message="Selected items have been bundled.",
        info={
            "items": [n for t in bundled_targets.targets for n in t.names],
        },
        echo=True,
    )



"""Shell data structures and functionality module."""
import argparse
import dataclasses
import datetime
import io
import shlex
import typing

import colorama
import prompt_toolkit
from prompt_toolkit.formatted_text import ANSI
from prompt_toolkit.history import InMemoryHistory

import reviser
from reviser import commands
from reviser import definitions
from reviser import templating
from reviser.interactivity import completions


@dataclasses.dataclass(frozen=True)
class ExecutionResult:
    """Data structure for shell command execution responses."""

    status: str
    message: str
    info: typing.Optional[dict] = None
    data: typing.Optional[dict] = None


@dataclasses.dataclass(frozen=True)
class Execution:
    """Data structure for a shell command execution."""

    action: str
    args: dict
    shell: "Shell"
    result: typing.Optional["ExecutionResult"] = None

    executed_at: datetime.datetime = dataclasses.field(
        init=False,
        default_factory=lambda: datetime.datetime.utcnow(),
    )

    @property
    def timestamp(self) -> str:
        """Get a string representation of the created at datetime."""
        return self.executed_at.strftime("%Y-%m-%dT%H:%M:%S.000Z")

    def finalize(
        self,
        status: str,
        message: str = None,
        info: dict = None,
        data: dict = None,
        echo: bool = False,
    ) -> "Execution":
        """Copy this execution and populate it with the result."""
        return dataclasses.replace(
            self,
            result=ExecutionResult(
                status=(status or "???").upper(),
                message=message or "???",
                info=info,
                data=data,
            ),
        ).echo_if(echo)

    def echo(self) -> "Execution":
        """Echo the result if set for display to the console."""
        if self.result:
            templating.printer(
                "interactivity/shells/execution_result.jinja2",
                result=self.result,
            )
        return self

    def echo_if(self, condition) -> "Execution":
        """Echo the result if set for display to the console."""
        if bool(condition):
            return self.echo()
        return self


class Shell:
    """
    Reviser Shell wrapper class.

    Interactive lambda deployer interactivity class responsible for managing
    the state and interactivity for the interface.
    """

    def __init__(
        self,
        context: "definitions.Context",
        selection: "definitions.Selection" = None,
    ):
        """Create a new shell for queued and/or interactive execution."""
        self.command_history: typing.List[str] = []
        self.execution_history: typing.List["Execution"] = []
        self._prompt_session: typing.Optional[prompt_toolkit.PromptSession] = None
        self.context = context
        self.selection = selection or definitions.Selection()
        self.shutdown = False
        #: Stores uncaught exceptions for reference when a command fails.
        self.error: typing.Optional[Exception] = None
        self._shell_completer = completions.ShellCompleter(self)
        self.command_queue: typing.List[str] = []
        #: Explicitly specifies if the shell should be running in
        #: interactive mode where it prompts users for input on the
        #: terminal.
        self._is_interactive: typing.Optional[bool] = None

    @property
    def is_interactive(self) -> typing.Optional[bool]:
        """Get whether the shell is running as an interactive command loop."""
        return self._is_interactive

    @is_interactive.setter
    def is_interactive(self, value: typing.Optional[bool]):
        """Set the is_interactive property."""
        if value and self._prompt_session is None:
            if not self._create_prompt_session():
                raise RuntimeError("Terminal connectivity not supported.")

        self._is_interactive = value

    @property
    def is_active(self) -> bool:
        """
        Get Whether the shell should continue running.

        This is determined by whether or not it is running interactively or there are
        queued commands to still process.
        """
        return not self.shutdown and bool(
            bool(self.command_queue) or self.is_interactive
        )

    def _create_prompt_session(self) -> bool:
        """
        Attempt creation of a session for interactive command execution.

        :return:
            Whether that session was successfully created. It will fail when executed
            in non-interactive environments that do no support user input and return
            False.
        """
        history = InMemoryHistory()
        for line in self.command_history:
            history.append_string(line)

        try:
            self._prompt_session = prompt_toolkit.PromptSession(history=history)
        except io.UnsupportedOperation:
            self._is_interactive = False
            return False

        return True

    def _setup(self):
        """Initialize before entering the command loop."""
        templating.printer(
            "interactivity/shells/splash.jinja2",
            version=reviser.__version__,
        )
        print("\n\n")
        colorama.init()

        # If no queued commands have been specified at the start of the
        # loop, consider the loop to be running in interactive mode.
        self.is_interactive = not bool(self.command_queue)

    def run(self) -> None:
        """Launch the command execution loop."""
        self._setup()

        try:
            while not self.shutdown and self.is_active:
                if line := self._get_next_command():
                    should_error_exit = execute(self, line)
                    self.shutdown = bool(self.shutdown or should_error_exit)

                if self.error and not self.is_interactive:
                    # Abort if an error was encountered in non-interactive mode.
                    return
        except KeyboardInterrupt:
            print("\n[INTERRUPTED]: Shutting down terminal.")
        except Exception as error:
            if self.error is None:
                # If the error was not already caught, catch it now.
                self.error = error
                templating.print_error(
                    error=error,
                    message="An unexpected command error occurred.",
                )
        finally:
            self._teardown()

    def _teardown(self):
        """Teardown after exiting the command loop."""
        print("\n\n")
        colorama.deinit()
        self.shutdown = True
        self.command_queue = []
        self.is_interactive = False
        self._prompt_session = None

    def _get_next_command(self) -> str:
        """
        Prompt user for input and returns that for command execution.

        However, if there is a queued command that is returned without a prompt instead.
        """
        context = self.context
        configuration = context.configuration
        templating.printer(
            "interactivity/shells/prompt.jinja2",
            profile=context.connection.session.profile_name or "default",
            region=configuration.aws_region,
            user_slug=context.connection.user_slug,
            selected=context.get_selected_targets(self.selection),
        )
        prompt = f"{colorama.Fore.GREEN}>{colorama.Style.RESET_ALL} "

        if self.command_queue:
            line = self.command_queue.pop(0)
            print(f"{prompt}{line}")
        elif self._prompt_session is not None:
            # noinspection PyTypeChecker
            line = self._prompt_session.prompt(
                message=ANSI(prompt),
                completer=self._shell_completer,
                complete_while_typing=True,
            )
        else:
            line = ""

        return line.strip()


def execute(shell: "Shell", line: str) -> bool:
    """
    Execute the specified input command within the given shell environment.

    :return:
        True if an error was encountered during execution and the shell should be
        shutdown as a consequence, which happens in non-interactive mode.
    """
    # Continue if interactive, but die if non-interactive.
    should_error_exit = not shell.is_interactive

    if len(line.strip()) < 1:
        return should_error_exit

    try:
        raw_args = shlex.split(line)
        action = raw_args[0]
        raw_args = raw_args[1:]
    except Exception as error:
        shell.error = error
        templating.print_error(
            error=error,
            message=f'An error occurred parsing "{line}".',
        )
        return should_error_exit

    action_module = commands.get_module(action)

    print("\n")
    if action_module is None:
        templating.print_error(f'Unknown command "{action}".')
        return should_error_exit

    try:
        try:
            parser = argparse.ArgumentParser(prog=action)
            action_module.populate_subparser(parser)
            args = vars(parser.parse_args(raw_args))
        except SystemExit:
            return should_error_exit

        execution = action_module.run(Execution(action, args, shell))
        shell.execution_history.append(execution)
        shell.command_history.append(line)
        return False
    except Exception as error:
        shell.error = error
        templating.print_error(
            error=error,
            message="An unexpected command error occurred.",
        )
        return should_error_exit
    finally:
        print("\n")



"""Invoke the target function."""
import json
import typing
from argparse import ArgumentParser

import yaml

from reviser import definitions
from reviser import interactivity
from reviser import templating


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return ["--function", "--response-format", "--tail"]


def populate_subparser(parser: ArgumentParser):
    """Add alias subcommand to supplied parser."""
    parser.add_argument(
        "--function",
        help="""
        The invoke command only acts on one function. This can be
        achieved either by selecting the function target via the
        select command, or specifying the function name to apply this
        change to with this flag.
        """,
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="""
        When specified, the invocation logging will be included as well for both the
        request and response data.
        """,
    )
    parser.add_argument(
        "--tail",
        action="store_true",
        help="""
        When specified, the tail of the log output from the function will be displayed
        as part of the execution. This will automatically be set if verbose is
        specified.
        """,
    )
    parser.add_argument(
        "--response-format",
        choices=["text", "json", "yaml"],
        default="text",
    )
    parser.add_argument(
        "--payload",
        default="{}",
        help="Specify the payload as a JSON-serialized string.",
    )
    parser.add_argument(
        "--payload-file",
        help="Specify the payload from a JSON/YAML file.",
    )


def _get_target_function(
    ex: "interactivity.Execution",
) -> typing.Tuple[typing.Optional["definitions.Target"], typing.Optional[str]]:
    """Get the target function that will be invoked."""
    selected = ex.shell.context.get_selected_targets(ex.shell.selection)
    function_name: typing.Optional[str] = ex.args.get("function")
    matches = [
        (target, name)
        for target in selected.function_targets
        for name in target.names
        if not function_name or name == function_name
    ]

    if not matches:
        ex.finalize(
            status="ERROR",
            message="No matching function targets found.",
            echo=True,
        )
        return None, None

    if len(matches) != 1:
        ex.finalize(
            status="ERROR",
            message="Invoke can only function on a single function.",
            echo=True,
        )
        return None, None

    return matches[0]


def _read_payload_file(ex: "interactivity.Execution") -> typing.Dict[str, typing.Any]:
    """Read invocation payload data from the file argument."""
    relative_file_path: typing.Optional[str] = ex.args.get("payload_file")
    if not relative_file_path:
        return {}

    path = ex.shell.context.configuration.directory.joinpath(relative_file_path)

    if path.name.endswith((".yaml", ".yml")):
        return yaml.safe_load(path.read_text())

    return json.loads(path.read_text())


def _get_payload(ex: "interactivity.Execution") -> typing.Dict[str, typing.Any]:
    """Assemble and return the payload for the function invocation."""
    value: typing.Dict[str, typing.Any] = json.loads(ex.args.get("payload", "{}"))
    value.update(_read_payload_file(ex))
    return value


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Execute the invoke command."""
    target, function_name = _get_target_function(ex)
    if not target:
        return ex

    payload = _get_payload(ex)
    is_verbose: bool = ex.args.get("verbose", False)
    tail_logs: bool = ex.args.get("tail", False) or is_verbose

    if is_verbose:
        print(
            "\n=== INVOCATION ===\n{}\n".format(
                yaml.safe_dump(
                    {
                        "function_name": function_name,
                        "payload": payload,
                    }
                )
            )
        )

    response = target.client("lambda").invoke(
        FunctionName=function_name,
        InvocationType="RequestResponse",
        LogType="Tail" if tail_logs else "None",
        Payload=json.dumps(payload).encode(),
    )

    if tail_logs and (logs := response.get("LogResult", "")):
        print(f"\n=== TAILED INVOCATION LOGS ===\n{logs}\n")

    invoked_version: str = response.get("ExecutedVersion", "$LATEST")

    if "FunctionError" in response:
        templating.print_error(response["FunctionError"])
        return ex.finalize(
            status="EXECUTION_ERROR",
            message="Function execution failed.",
            data={
                "version": invoked_version,
                "function_error": response.get("FunctionError"),
                "status_code": response.get("StatusCode"),
                "payload": payload,
            },
        )

    return ex.finalize(
        status="SUCCESS",
        message="Function invoked successfully.",
        data={
            "version": invoked_version,
            "payload": payload,
        },
    )



"""Publisher functionality module."""
import typing

from botocore.client import BaseClient

from reviser import definitions
from ..deploying import updater


def _wait_for_existing_updates_to_complete(client: BaseClient, function_name: str):
    """
    Wait for any existing updates to complete on the lambda function.

    This is used to make sure the function is ready to be updated.

    :param client:
        Lambda boto3 client
    :param function_name:
        Lambda function name
    :param waiter_type:
        The client.get_waiter type.
    """
    waiter = client.get_waiter("function_updated")
    waiter.wait(FunctionName=function_name)


def _update_function_configuration(
    client: BaseClient,
    function_name: str,
    target: "definitions.Target",
    published_layers: typing.List["definitions.PublishedLayer"],
    dry_run: bool,
):
    """
    Update the function configuration.

    Functions can only be updated if they are ready to be updated, so we will
    have to wait for the function State to be Active, and LastUpdateStatus to be
    Successful. It will exponentially increment the sleep time for every
    status check.

    :param client:
        Lambda boto3 client
    :param function_name:
        Lambda function name
    :param target:
        The definitions.Target.
    :param published_layers:
        The definitions.PublishedLayer
    """
    _wait_for_existing_updates_to_complete(client, function_name)
    updater.update_function_configuration(
        function_name=function_name,
        target=target,
        published_layers=published_layers,
        dry_run=dry_run,
    )


def _publish_function_version(
    client: BaseClient, function_name: str, code_sha_256: str, description: str
):
    """
    Publish function new version, waiting for existing updates to complete.

    Functions can only be updated if they are ready to be updated, so we will
    have to wait for the function State to be Active, and LastUpdateStatus to be
    Successful. It will exponentially increment the sleep time for every status
    check.

    :param client:
        Lambda boto3 client
    :param function_name:
        Lambda function name
    :param code_sha_256:
        The CodeSha256 from the client.update_function_cod response.
    :param description:
        The publish description
    :return:
        The client.publish_version response
    """
    _wait_for_existing_updates_to_complete(client, function_name)
    return client.publish_version(
        FunctionName=function_name,
        CodeSha256=code_sha_256,
        Description=description or "",
    )


def publish_function(
    target: "definitions.Target",
    s3_keys: typing.Optional[typing.List[str]] = None,
    published_layers: typing.Optional[typing.List["definitions.PublishedLayer"]] = None,
    description: str = None,
    dry_run: bool = False,
):
    """
    Publish an updated version of the lambda function.

    :param target:
        The lambda function to publish.
    :param s3_keys:
        If the lambda function is not image based, the S3 keys of the code
        artifacts to deploy. If the lambda function is image based this will
        be ignored. This is expected to be an ordered list aligning with the
        target's names.
    :param published_layers:
        The published lambda layers to connect to the lambda function.
    :param description:
        A description of the lambda function version.
    :param dry_run:
        Whether to actually perform the action.
    """
    s3_keys = s3_keys or []
    published_layers = published_layers or []
    client = target.client("lambda")
    for i, name in enumerate(target.names):
        print(f"[PUBLISHING]: Deploying update to {name} $LATEST")
        response = None
        s3_key = s3_keys[i] if len(s3_keys) > i else None
        image = target.image.get_region_uri(target.aws_region)

        if not dry_run:
            code_args = (
                typing.cast(typing.Dict[str, str], {"ImageUri": image})
                if image
                else typing.cast(
                    typing.Dict[str, str], {"S3Bucket": target.bucket, "S3Key": s3_key}
                )
            )
            response = client.update_function_code(
                FunctionName=name, Publish=False, **code_args
            )

        _update_function_configuration(
            client=client,
            function_name=name,
            target=target,
            published_layers=published_layers,
            dry_run=dry_run,
        )

        print("[PUBLISHING]: Publishing new version")
        if response and not dry_run:
            response = _publish_function_version(
                client=client,
                function_name=response["FunctionName"],
                code_sha_256=response["CodeSha256"],
                description=description or "",
            )
            print(
                "[PUBLISHED]: Function {} ({})".format(
                    name,
                    response["Version"],
                )
            )
            print("  - Version:", response["FunctionArn"])

    print("[DEPLOYED]: Lambda function code has been deployed\n")


def publish_layer(
    target: "definitions.Target",
    s3_keys: typing.List[str],
    description: str = None,
    dry_run: bool = False,
) -> typing.List["definitions.PublishedLayer"]:
    """Publish an updated version of the layer after bundle has been uploaded to S3."""
    published_layers = []
    client = target.client("lambda")

    for name, key in zip(target.names, s3_keys):
        print(f"[PUBLISHING]: Publishing code bundle to {name} layer")
        if dry_run:
            continue

        response = client.publish_layer_version(
            LayerName=name,
            Description=description or "",
            Content={
                "S3Bucket": target.bucket,
                "S3Key": key,
            },
            CompatibleRuntimes=[f"python{definitions.RUNTIME_VERSION}"],
        )
        print("[PUBLISHED]: Layer {} ({})".format(name, response["Version"]))
        print("  - Layer:", response["LayerArn"])
        print("  - Version:", response["LayerVersionArn"])
        published_layers.append(definitions.PublishedLayer(response=response))

    print("[DEPLOYED]: Lambda layer code has been deployed\n")
    return published_layers



"""Lambda layer functionality module."""
import textwrap
import typing
from collections import defaultdict

from botocore.client import BaseClient

from reviser import definitions
from reviser import templating
from reviser.servicer import functioning


def get_layer_versions(
    lambda_client: BaseClient,
    layer_name: str,
) -> typing.List["definitions.LambdaLayer"]:
    """
    Get a list of layer versions for the specified layer name (or ARN).

    These are sorted increasing version order.

    :param lambda_client:
        Boto3 client used to query for layer version info.
    :param layer_name:
        Name or ARN of the layer for which to list versions.
    """
    try:
        paginator = lambda_client.get_paginator("list_layer_versions")
        request = {"LayerName": layer_name}
        layers = [
            definitions.LambdaLayer(layer)
            for page in paginator.paginate(**request)
            for layer in page.get("LayerVersions") or []
        ]
        return list(sorted(layers, key=lambda x: x.version or 0))
    except lambda_client.exceptions.ResourceNotFoundException:
        return []


def remove_layer_version(lambda_client: BaseClient, layer_arn: str) -> bool:
    """Remove the specified lambda layer."""
    parts = layer_arn.rsplit(":", 1)
    request = dict(
        LayerName=parts[0],
        VersionNumber=int(parts[1]),
    )
    try:
        lambda_client.delete_layer_version(**request)
        return True
    except Exception as error:
        print("REQUEST:", request)
        templating.print_error(
            f'[FAILED]: Version {layer_arn} could not be deleted "{error}"', error
        )
        return False


def get_layer_version(
    lambda_client: BaseClient,
    layer_name: str,
    version: int,
) -> "definitions.LambdaLayer":
    """Retrieve the configuration for the specified lambda layer."""
    return definitions.LambdaLayer(
        lambda_client.get_layer_version(
            LayerName=layer_name,
            VersionNumber=version,
        )
    )


def _get_associated_function_versions(
    client: BaseClient,
    function_names: typing.List[str],
) -> typing.List["definitions.LambdaFunction"]:
    return [
        version
        for name in function_names
        for version in functioning.get_function_versions(client, name)
    ]


def _get_attached_function_versions(
    layer_name: str,
    function_versions: typing.List["definitions.LambdaFunction"],
) -> typing.DefaultDict[str, typing.List[str]]:
    """Get a list of function version and alias attachments for this layer."""
    functions: typing.DefaultDict[str, typing.List[str]] = defaultdict(list)
    for func in function_versions:
        match = func.get_layer(layer_name)
        if match is not None:
            functions[str(match.version)] += [
                f"{func.name}:{func.version}",
                *[f"{func.name}:{a.name}" for a in func.aliases],
            ]
    return functions


def echo_layer_versions(
    client: BaseClient,
    layer_names: typing.List[str],
    function_names: typing.List[str],
):
    """Print a table of lambda layer versions to the terminal."""
    function_versions = _get_associated_function_versions(client, function_names)

    for name in layer_names:
        versions = get_layer_versions(client, name)
        if not versions:
            print(f'\n[IGNORED]: No layer "{name}" was found.')
            continue

        functions = _get_attached_function_versions(name, function_versions)

        print("\n{}".format(64 * "="))
        print(f"(L) {name}")
        for v in versions:
            print(
                textwrap.indent(
                    "\n".join(
                        textwrap.wrap(
                            f"{v.version}: {v.description or v.created}",
                            width=60,
                            subsequent_indent="  ",
                        )
                    ),
                    prefix="  ",
                )
            )
            for func in functions[str(v.version)]:
                print(f"      - {func}")



"""Context definitions module."""
import argparse
import dataclasses
import pathlib
import typing

import yaml

from reviser.definitions import aws
from reviser.definitions import configurations
from reviser.definitions import enumerations
from reviser.definitions import selections


@dataclasses.dataclass(frozen=True)
class SelectedTargets:
    """
    Data structure that acts as a container for the selected targets.

    this container is for selected target results that have been filtered down from the
    entire list of targets stored in the parent configuration.
    """

    targets: typing.List["configurations.Target"]

    @property
    def function_targets(self) -> typing.List["configurations.Target"]:
        """List function targets in the selection."""
        return [
            t
            for t in (self.targets or [])
            if t.kind == enumerations.TargetType.FUNCTION
        ]

    @property
    def layer_targets(self) -> typing.List["configurations.Target"]:
        """List layer targets in the selection."""
        return [
            t for t in (self.targets or []) if t.kind == enumerations.TargetType.LAYER
        ]


@dataclasses.dataclass(frozen=True)
class Context:
    """Execution context for the current invocation."""

    arguments: argparse.Namespace
    configuration: "configurations.Configuration"
    connection: "aws.AwsConnection"

    @property
    def command_queue(self) -> typing.Optional[typing.List[str]]:
        """
        List commands to process within the execution shell.

        These commands will execute and then exit instead of running interactively.
        These are loaded from `run:` configuration command group definitions in the
        loaded configuration object if a `--run` argument specifies the command group
        definition to execute. Otherwise, this will be None.
        """
        name = self.arguments.command_group_name
        return self.configuration.get_as_list("run", name) or None

    def get_selected_targets(
        self, selection: "selections.Selection"
    ) -> "SelectedTargets":
        """Create a target selection from the specified selection criteria."""
        return SelectedTargets(
            targets=[
                ts
                for t in self.configuration.targets
                if (ts := dataclasses.replace(t, selection=selection)).names
            ]
        )

    @classmethod
    def load_from_file(
        cls,
        arguments: argparse.Namespace,
        path: typing.Union[str, pathlib.Path] = None,
        connection: "aws.AwsConnection" = None,
    ) -> "Context":
        """
        Load a context from a configuration path target.

        Loads files using the `lambda.yaml` file by default as this is the preferred
        extension: https://yaml.org/faq.html
        """
        target = pathlib.Path(path or "./lambda.yaml").absolute()
        if target.is_dir():
            target = target.joinpath("lambda.yaml")

        aws_connection = connection or aws.AwsConnection()
        contents = target.read_text()
        return cls(
            arguments=arguments,
            configuration=configurations.Configuration(
                directory=target.parent,
                data=yaml.safe_load(contents),
                connection=aws_connection,
            ),
            connection=aws_connection,
        )



"""Subpackage for CLI display templating."""
import pathlib
import re
import textwrap
import traceback
import typing

import colorama
import yaml
from jinja2 import Environment

import reviser

_COLOR_MAP = {
    "green": colorama.Fore.GREEN,
    "red": colorama.Fore.RED,
    "error": colorama.Fore.RED,
    "cyan": colorama.Fore.CYAN,
    "magenta": colorama.Fore.MAGENTA,
    "yellow": colorama.Fore.YELLOW,
    "warning": colorama.Fore.YELLOW,
}


def _yaml_filter(value: dict, indent: int = 0) -> str:
    """Render the value as yaml."""
    if not value:
        return ""

    return textwrap.indent(yaml.safe_dump(value), prefix=indent * " ")


def _single_line_filter(value: str) -> str:
    """Reduce a string to a single line with no insignificant padding."""
    regex = re.compile(r"\s{2,}")
    return regex.sub(" ", (value or "").replace("\n", " ")).strip()


def _colorize_filter(value: str, color: str = None) -> str:
    """Apply ansi console coloring to the string."""
    if color_value := _COLOR_MAP.get((color or "").lower()):
        return f'{color_value}{value or ""}{colorama.Style.RESET_ALL}'
    return value


_environment = Environment()
_environment.filters["yaml"] = _yaml_filter
_environment.filters["single_line"] = _single_line_filter
_environment.filters["colorize"] = _colorize_filter


def render(location: str, **kwargs) -> str:
    """Load a template from the path and returns the rendered contents."""
    contents = (
        pathlib.Path(reviser.__file__)
        .parent.joinpath(*location.strip("/").split("/"))
        .read_text()
    )
    return _environment.from_string(contents).render(**kwargs)


def printer(location: str, **kwargs):
    """Render the template and prints it to stdout."""
    print(render(location, **kwargs))


def print_error(message: str, error: Exception = None):
    """
    Render an exception template to stdout.

    If error is specified, that will be rendered as a stacktrace as well.
    """
    stack_trace: typing.Optional[str]
    if error:
        stack_trace = "\n".join(
            traceback.format_exception(
                type(error),
                error,
                error.__traceback__,
            )
        )
    else:
        stack_trace = None
    printer(
        "templating/error.jinja2",
        message=message,
        stack_trace=stack_trace,
    )



"""Shell interactivity subpackge."""
from reviser.interactivity.completions import ShellCompleter  # noqa: F401
from reviser.interactivity.runner import create_shell  # noqa: F401
from reviser.interactivity.runner import main_shell  # noqa: F401
from reviser.interactivity.runner import run_shell  # noqa: F401
from reviser.interactivity.shells import Execution  # noqa: F401
from reviser.interactivity.shells import ExecutionResult  # noqa: F401
from reviser.interactivity.shells import Shell  # noqa: F401



"""Switch the target region."""
import argparse
import typing

from reviser import interactivity

# https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Concepts.RegionsAndAvailabilityZones.html  # noqa
AVAILABLE_REGIONS = {
    "us-east-2": "US East (Ohio)",
    "us-east-1": "US East (N. Virginia)",
    "us-west-1": "US West (N. California)",
    "us-west-2": "US West (Oregon)",
    "af-south-1": "Africa (Cape Town)",
    "ap-east-1": "Asia Pacific (Hong Kong)",
    "ap-south-1": "Asia Pacific (Mumbai)",
    "ap-northeast-3": "Asia Pacific (Osaka-Local)",
    "ap-northeast-2": "Asia Pacific (Seoul)",
    "ap-southeast-1": "Asia Pacific (Singapore)",
    "ap-southeast-2": "Asia Pacific (Sydney)",
    "ap-northeast-1": "Asia Pacific (Tokyo)",
    "ca-central-1": "Canada (Central)",
    "cn-north-1": "China (Beijing)",
    "cn-northwest-1": "China (Ningxia)",
    "eu-central-1": "Europe (Frankfurt)",
    "eu-west-1": "Europe (Ireland)",
    "eu-west-2": "Europe (London)",
    "eu-south-1": "Europe (Milan)",
    "eu-west-3": "Europe (Paris)",
    "eu-north-1": "Europe (Stockholm)",
    "me-south-1": "Middle East (Bahrain)",
    "sa-east-1": "South America (São Paulo)",
    "us-gov-east-1": "AWS GovCloud (US-East)",
    "us-gov-west-1": "AWS GovCloud (US-West)",
}


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return list(AVAILABLE_REGIONS.keys())


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    parser.add_argument(
        "region_name",
        choices=list(AVAILABLE_REGIONS.keys()),
        nargs="?",
        help="""
            AWS region name for the override. Leave it blank to return to
            the default region for the initially loaded credentials and/or
            environment variables.
            """,
    )


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Switch the configured region for the shell."""
    region_name = ex.args.get("region_name")
    configs = ex.shell.context.configuration

    if not region_name and "region" in configs.data:
        del configs.data["region"]
    elif region_name:
        configs.data["region"] = region_name

    return ex.finalize(
        status="SWITCHED_REGION",
        message=f"Now in region {configs.aws_region}.",
        echo=True,
        data={"region": configs.aws_region},
    )



"""Commands subpackage for reviser command invcations."""
from collections import defaultdict as _defaultdict

from ..commands import aliaser
from ..commands import bundler
from ..commands import configer
from ..commands import deployer
from ..commands import exiter
from ..commands import helper
from ..commands import lister
from ..commands import pruner
from ..commands import pusher
from ..commands import region_switcher
from ..commands import reloader
from ..commands import reporter
from ..commands import selector
from ..commands import sheller
from ..commands import tailer

#: Commands available to the shell.
COMMANDS = {
    "alias": aliaser,
    "bundle": bundler,
    "configs": configer,
    "deploy": deployer,
    "exit": exiter,
    "help": helper,
    "list": lister,
    "prune": pruner,
    "push": pusher,
    "region": region_switcher,
    "reload": reloader,
    "select": selector,
    "shell": sheller,
    "status": reporter,
    "tail": tailer,
}

#: Aliases of commands to make available to the shell.
ALIASES = {
    "?": "help",
}


def _invert_aliases():
    """Reverse the alias dictionary to be a lookup from command to aliases."""
    out = _defaultdict(list)
    for alias, name in ALIASES.items():
        out[name].append(alias)
    return out


#: List of aliases assigned to shell commands for documentation purposes.
REVERSED_ALIASES = _invert_aliases()


def get_module(command: str):
    """Retrieve the command module for the associated command."""
    return COMMANDS.get(ALIASES.get(command, command), None)



"""
Macro command to convert to interactive shell operation.

This is a special command to use in run command groups/macros to start interactive
command mode for the terminal. Useful when in scenarios where you wish to prefix an
interactive session with commonly executed commands. For example, if you want to select
certain targets with the select command as part of starting the shell, you could create
a run command group/macro in your lambda.yaml that executes the select command and then
executes the shell command. This would updated the selection and then with the shell
command, start the shell in interactive mode. Without specifying the shell command
here, the run command group/macro would just set a selection and then exit.
"""
import argparse
import typing

from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return []


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    pass


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Switch to an interactive shell."""
    ex.shell.is_interactive = True
    return ex.finalize(
        status="SHELL",
        message="Switching to interactive shell execution.",
        echo=True,
    )



"""Subpackage for assembling a lambda code package bundle for upload and deployment."""
import os
import zipfile

from reviser import definitions
from reviser.bundling import _installer


def _create_zip(target: "definitions.Target"):
    """
    Bundle together all of the source files into a single zip file.

    This zip file is structured so that it can be uploaded as an AWS Lambda function.

    :param target:
        A target configuration to bundle into a zip file for deployment.
    """
    zip_bundle = zipfile.ZipFile(
        target.bundle_zip_path,
        mode="w",
        compression=zipfile.ZIP_DEFLATED,
        compresslevel=9,
    )

    for site_package_path in target.bundle.get_site_package_paths():
        # Add installed site package files into the bundle. For functions
        # the site packages should be added to the top-level bundle directory
        # as is the convention recommended by AWS documentation. However,
        # for layers, the site packages folder should be preserved during
        # bundling, which is why the root folders below are different for
        # the two types.
        root = (
            target.site_packages_directory
            if target.kind == definitions.TargetType.FUNCTION
            else target.bundle_directory
        )
        name = str(site_package_path.relative_to(root))
        zip_bundle.write(str(site_package_path), arcname=name)

    for copy_path in target.bundle.get_copy_paths():
        # Add copied source files into the bundle.
        name = str(copy_path.destination.relative_to(target.bundle_directory))
        zip_bundle.write(str(copy_path.destination), arcname=name)

    zip_bundle.close()
    print(f"[ARCHIVED]: {target.bundle_zip_path}")


def create(
    context: "definitions.Context",
    selection: "definitions.Selection",
    reinstall: bool = False,
):
    """
    Assemble the contents of the bundle folder for zipping.

    This copies the entire contents of the project folder to a temporary location
    and then installs and bundles the copied project for deployment.

    :param context:
        The context in which this bundle action is taking place.
    :param selection:
        A selection object that defines the subset of targets to execute
        the bundling process upon.
    :param reinstall:
        Whether or not to force re-installation of dependencies for previously
        bundled targets.
    """
    selected = context.get_selected_targets(selection)

    for target in selected.targets:
        target.bundle_directory.mkdir(exist_ok=True, parents=True)

        skip_installs = (
            not reinstall
            and target.site_packages_directory.exists()
            and os.listdir(str(target.site_packages_directory))
        )
        if not skip_installs:
            _installer.install_dependencies(target)
        else:
            print("[DEPENDENCIES]: Using existing installation cache.")

        for copy_path in target.bundle.get_copy_paths():
            copy_path.copy()
            print(f"[COPIED]: {copy_path.source.relative_to(target.directory)}")

        _create_zip(target)

    return selected



"""Shell execution module."""
import pathlib
import typing
import sys

import boto3

from reviser import definitions
from reviser import interactivity
from reviser import parsing


def create_shell(arguments: typing.List[str] = None) -> "interactivity.Shell":
    """
    Create a shell for interactive or queued command execution.

    The type of shell is based on the specified command line arguments, which default
    to sys.argv if no arguments are specified.
    """
    args = parsing.create_parser(True).parse_args(arguments)
    directory = pathlib.Path(args.root_directory).absolute()
    session = boto3.Session(
        profile_name=args.aws_profile_name,
        region_name=args.aws_region_name,
    )
    context = definitions.Context.load_from_file(
        arguments=args,
        path=directory,
        connection=definitions.AwsConnection(session),
    )
    return interactivity.Shell(context)


def run_shell(
    arguments: typing.List[str] = None,
    command_queue: typing.List[str] = None,
) -> "interactivity.Shell":
    """
    Start the shell process.

    The shell is started after first loading configuration from the target
    directory determined by the command line arguments.
    """
    shell = create_shell(arguments)
    shell.command_queue += command_queue or shell.context.command_queue or []
    shell.run()
    return shell


def main_shell() -> None:  # pragma: no cover
    """Execute entrypoint for the local shell cli."""
    shell = run_shell()
    sys.exit(1 if shell.error else 0)



"""Shared utility function module."""
import re
import typing


PACKAGE_NAME_REGEX = re.compile(r"^(?P<name>[^><=]+)")


def to_human_readable_size(
    number_of_bytes: typing.Optional[int],
) -> typing.Optional[str]:
    """Convert bytes value to human readable size."""
    if number_of_bytes is None:
        return None

    step_to_greater_unit = 1024
    unit = "bytes"
    value = float(number_of_bytes)

    if value < 1024:
        return f"{int(value)} {unit}"

    if (value / step_to_greater_unit) >= 1:
        value /= step_to_greater_unit
        unit = "KB"

    if (value / step_to_greater_unit) >= 1:
        value /= step_to_greater_unit
        unit = "MB"

    if (value / step_to_greater_unit) >= 1:
        value /= step_to_greater_unit
        unit = "GB"

    if (value / step_to_greater_unit) >= 1:
        value /= step_to_greater_unit
        unit = "TB"

    value = round(value, 1)
    return f"{value} {unit}"


def get_matching_bucket(
    buckets: typing.Union[None, str, dict],
    aws_account_id: str,
    aws_region: str,
    default_bucket: str = None,
):
    """Fetch a bucket name from the buckets argument."""
    if not buckets:
        return default_bucket

    if isinstance(buckets, str):
        return buckets

    # Allow for setting buckets by account ID.
    account_buckets = buckets.get(aws_account_id, buckets)

    if isinstance(account_buckets, str):
        return account_buckets

    # Allow for settings buckets by aws region name.
    return account_buckets.get(aws_region, default_bucket)


def extract_package_name(package_identifier: str) -> str:
    """Extract the name from a package specification, which can include a version."""
    match = PACKAGE_NAME_REGEX.search(package_identifier or "")
    return match.group("name") if match else ""
